import collections

postags = collections.defaultdict(int)
chunktags = collections.defaultdict(int)
btags = collections.defaultdict(int)
itags = collections.defaultdict(int)
identitylabels = collections.defaultdict(int)
worddictionary = collections.defaultdict(int)
sentencecount = 0
wordcount = 0

features = []
labels = []

# # dictionry for label-index translation
backtransdict = collections.defaultdict(int)

# # for preprocessing, filter all certain O's

Opostags = []
Ochunktags = []
Owords = []

###################### support methods ############################

def splitsentence(filename):
  sentencearray = []
  for line in open(str(filename)):
  	taglist = [x for x in line.split()]
 	sentencearray.append(taglist)
  return sentencearray

def splitsencetest(filename):
	sentencearray = []
	for line in open(str(filename)):
		taglist = [x for x in line.split()]
		sentencearray.append(taglist)
	return sentencearray

# # dictionary innerproduct
def innerproduct(weightvector,predictvector,features):
	dotp = 0
	for item in features:
		dotp+=weightvector[item]*predictvector[item]
	return dotp

# # vector add and substract
def vectoradd(vector1,vector2):
	for index,item in enumerate(vector1):
		vector1[item]+=vector2[item]

def vectorsubstract(vector1,vector2):
	for index,item in enumerate(vector1):
		vector1[item]-=vector2[item]

def avgmax(weightvectorarray,predictvector):
	y = -99999999
	amax = 0
	for index,item in enumerate(weightvectorarray):
		innerprod = innerproduct(weightvectorarray[index],predictvector,features)
		if innerprod>=y:
			y = innerprod
			amax = index
	return amax

def makebetter(weightvectorarray,wordvector,correctarg,wrongarg):
	vectoradd(weightvectorarray[correctarg],wordvector)
	vectorsubstract(weightvectorarray[wrongarg],wordvector)

# # build a list of weight vectors, index is same as labels, 
# # each index points to a dictionary of {featurename: value} pairs
def buildweightvectors(features,labels):
	numfeatures = len(features)
	numlables = len(labels)
	weightvectorarray = [0 for x in range(numlables)]
	for index,label in enumerate(labels):
		weightvectorarray[index] = collections.defaultdict(int)
		for item in features:
			weightvectorarray[index][item] = 1
	return weightvectorarray

# # build a defauilt feature vector for a word
# # position will be updated outside of this loop
# # incorporate word as features?
def buildfeaturedict(features,sentencelen,wordtaglist,position,prewordtaglist,nextwordtaglist):
	fdict = collections.defaultdict(int)
	for item in features:
		fdict[item] = 0
	if position is 0:
		fdict['sentencehead'] = 3
	if prewordtaglist!=[]:
		fdict[prewordtaglist[1]]+=1
	if nextwordtaglist!=[]:
		fdict[nextwordtaglist[1]]+=1
	fdict[wordtaglist[2]]+=2
	fdict[wordtaglist[1]]+=3
	return fdict

# # split array for 10-fold testing
def traintestprocess(fullarray,position1,position2):
  lenarray = len(fullarray)
  trainlist1 = fullarray[0:position1*lenarray/10]
  trainlist2 = fullarray[position2*lenarray/10:]
  if trainlist1[0] == []:
  	del trainlist1[0]
  if len(trainlist1)>0 and trainlist1[len(trainlist1)-1]!=[]:
  	trainlist1.append([])
  if trainlist2[0] == []:
  	del trainlist2[0]
  if len(trainlist2)>0 and trainlist2[len(trainlist2)-1]!=[]:
  	trainlist1.append([])
  trainlist = trainlist1 + trainlist2
  testlist = fullarray[position1*lenarray/10:position2*lenarray/10]
  if testlist[0] == []:
  	del testlist[0]
  if len(testlist)>0 and testlist[len(testlist)-1]!=[]:
  	testlist.append([])
  return [trainlist,testlist]

###################################################################

def blacklistpostags(sentencearray,postags):
	blacklistcandidates = collections.defaultdict(float)
	for item in postags:
		blacklistcandidates[item] = 0
	for item in sentencearray:
		if item!=[] and item[3]!='O':
			blacklistcandidates[item[1]]+=1
	for item in blacklistcandidates:
		blacklistcandidates[item]=blacklistcandidates[item]/(postags[item]*1.0)
	realblacklist = [x for x in blacklistcandidates if blacklistcandidates[x]<0.1]
	return realblacklist

def preprocess(sentencearray):
	OWordArray = []
	for item in sentencearray:
		if item!=[] and item[3].strip()=='O' and (item[0] not in OWordArray):
			OWordArray.append(item[0])
	return OWordArray

def avgperceptron(wvectorarray,sarray,farray,btdict,labels,oarray,blacklist):
	position = 0
	sentencelen = 0
	for i,item in enumerate(sarray):
		# # construct sentence length (sentencelen in function buildfeaturedict)
		# # if current position is at start
		if position is 0:
			sentencelen = 0
			while (sarray[i+sentencelen] != []):
				sentencelen+=1
		if item != []:
			if item[0] in oarray:
				amax = 2
			elif item[1] in blacklist:
				amax = 2
			else:
				pretaglist = []
				nexttaglist = []
				if position!=0 and position!=sentencelen-1 and sentencelen>2:
					pretaglist = sarray[i-1]
					nexttaglist = sarray[i+1]
				wordfeaturevector = buildfeaturedict(farray,sentencelen,sarray[i],position,pretaglist,nexttaglist)
				# # train weight vectors
				amax = avgmax(wvectorarray,wordfeaturevector)
				if sarray[i][3] != labels[amax]:
					makebetter(wvectorarray,wordfeaturevector,btdict[sarray[i][3]],amax)
			########################
			position+=1
		else:
			position = 0
	return wvectorarray

# # prediction function in 10-fold test
def testperceptron(wvectorarray,sarray,farray,btdict,labels,oarray,blacklist):
	position = 0
	sentencelen = 0
	count = 0
	hit = 0
	test = 0
	testrate = 0
	for i,item in enumerate(sarray):
		# # construct sentence length (sentencelen in function buildfeaturedict)
		# # if current position is at start
		if position is 0:
			sentencelen = 0
			while (sarray[i+sentencelen] != []):
				sentencelen+=1
		if item != []:
			if item[0] in oarray:
				amax = 2
			elif item[1] in blacklist:
				amax = 2
			else:
				pretaglist = []
				nexttaglist = []
				if position!=0 and position!=sentencelen-1 and sentencelen>2:
					pretaglist = sarray[i-1]
					nexttaglist = sarray[i+1]
				wordfeaturevector = buildfeaturedict(farray,sentencelen,sarray[i],position,pretaglist,nexttaglist)
				amax = avgmax(wvectorarray,wordfeaturevector)
			if sarray[i][2][0] == 'B' and labels[amax]!='O' and labels[amax][0]=='I':
				testrate+=1
			if sarray[i][2][0] == 'I' and labels[amax]!='O' and labels[amax][0]=='B':
				testrate+=1
			if sarray[i][3] == labels[amax]:
				hit+=1
			elif (sarray[i][3][0]=='B' and labels[amax][0]=='I') or (sarray[i][3][0]=='I' and labels[amax][0]=='B'):
				test+=1
			position+=1
		else:
			position = 0
		count+=1
	print count
	print hit
	print (hit+testrate)/(count*1.0)
	print test

# # read test file


# # prediction function in 10-fold test
def perceptronprediction(wvectorarray,sarray,farray,btdict,labels,oarray,blacklist):
	position = 0
	sentencelen = 0
	count = 0
	hit = 0
	test = 0
	testrate = 0
	for i,item in enumerate(sarray):
		# # construct sentence length (sentencelen in function buildfeaturedict)
		# # if current position is at start
		if item[0]=="<s>":
			print 'O'
			position = 0
			continue
		if position is 0:
			sentencelen = 0
			while (sarray[i+sentencelen][0] != "."):
				sentencelen+=1
			sentencelen+=1
		if item != []:
			amax = 2
			if item[0] in oarray:
				amax = 2
				print 'O'
				continue
			elif item[1] in blacklist:
				amax = 2
				print 'O'
				continue
			else:
				pretaglist = []
				nexttaglist = []
				if position!=0 and position!=sentencelen-1 and sentencelen>2:
					pretaglist = sarray[i-1]
					nexttaglist = sarray[i+1]
				wordfeaturevector = buildfeaturedict(farray,sentencelen,sarray[i],position,pretaglist,nexttaglist)
				amax = avgmax(wvectorarray,wordfeaturevector)
			if sarray[i][2][0] == 'B' and labels[amax]!='O' and labels[amax][0]=='I':
				testrate+=1
				print 'B' + labels[amax][1::]
				continue
			if sarray[i][2][0] == 'I' and labels[amax]!='O' and labels[amax][0]=='B':
				testrate+=1
				print 'I' + labels[amax][1::]
				continue
			print labels[amax]
			position+=1
		else:
			position = 0
		count+=1

###################################################################

sentencearray = splitsentence("english.train")
sentencearray.append([])

# # do stats for pos tags, chunk tags, labels, etc.

stemp = 0
for item in sentencearray:
	if len(item) is not 0:
		wordcount+=1
		worddictionary[item[0]]+=1
		postags[item[1]]+=1
		chunktags[item[2]]+=1
		identitylabels[item[3]]+=1
		if stemp is 0:
			sentencecount+=1
			stemp = 1
	else:
		stemp = 0

# # build feature vectors

for key in postags.keys():
	features.append(key)
	features.append('pretag'+key)
	features.append('nexttag'+key)
for key in chunktags.keys():
	features.append(key)
features.append('sentencehead')
#features.append('sentenceend')

# # get all labels

for key in identitylabels.keys():
	labels.append(key)

# # fill out backtrans dict
for index,item in enumerate(labels):
	backtransdict[item] = index

# # build weight vectors, default to 0

blacklist = blacklistpostags(sentencearray,postags)
blacklist.append("<s>")

for item in blacklist:
	filter(lambda a: a != item, features)

weightvectorarray = buildweightvectors(features,labels)

# # build train vectors for each sentence
# # keep track of sentence position in sentence array
# # train the model (weightvectorarray)

# position = 0
# sentencelen = 0
# for i,item in enumerate(sentencearray):
# 	# # construct sentence length (sentencelen in function buildfeaturedict)
# 	# # if current position is at start
# 	if position is 0:
# 		sentencelen = 0
# 		while (sentencearray[i+sentencelen] != []):
# 			sentencelen+=1
# 	if item != []:
# 		wordfeaturevector = buildfeaturedict(features,sentencelen,sentencearray[i],position)

# 		# # train weight vectors

# 		amax = avgmax(weightvectorarray,wordfeaturevector)
# 		if sentencearray[i][3] != backtransdict[amax]:
# 			makebetter(weightvectorarray,wordfeaturevector,backtransdict[sentencearray[i][3]],amax)
# 		########################

# 		position+=1
# 	else:
# 		position = 0

# avgperceptron(weightvectorarray,sentencearray,features,backtransdict)
# print weightvectorarray

OWords = []
OWords.append("<s>")
OWords = preprocess(sentencearray)

####################################################
# # 10-fold testing

# traintestarray = traintestprocess(sentencearray,6,7)

# # # # testing

# if(traintestarray[0][0]==[]):
# 	del traintestarray[0][0]
# if(traintestarray[0][len(traintestarray[0])-1]!=[]):
# 	traintestarray[0].append([])
# if(traintestarray[1][0]==[]):
# 	del traintestarray[1][0]
# if(traintestarray[1][len(traintestarray[1])-1]!=[]):
# 	traintestarray[0].append([])

# avgperceptron(weightvectorarray,traintestarray[0],features,backtransdict,labels,OWords,blacklist)
# perceptronprediction(weightvectorarray,traintestarray[1],features,backtransdict,labels,OWords,blacklist)

###################################################################

# # build test vectors

testsentencearray = splitsencetest('english.test')

# # test

avgperceptron(weightvectorarray,sentencearray,features,backtransdict,labels,OWords,blacklist)
perceptronprediction(weightvectorarray,testsentencearray,features,backtransdict,labels,OWords,blacklist)

###################################################################

